import pandas as pd
import numpy as np
import os
import argparse
from rdkit import Chem
from rdkit.Chem import AllChem
import subprocess

def AddCM5Charges(mol,qcm5): 
    qcm5.to_csv('CM5_charges.csv',index=False,float_format='%6.4f')
    cm1a = mol.MolData['Q_LJ'] 
    cm5_charges = list(qcm5.CM5_final)
    assert len(cm1a) == len(qcm5), 'Please check the at_info and Q_LJ_dat files'
    for c in range(len(cm1a)): 
        cm1a[c][1] = '%9.6f'%cm5_charges[c]
    mol.MolData['Q_LJ'] = cm1a
    return mol 

def AtomFPProgram(hmol,atomNum,radii=2):
    env = Chem.FindAtomEnvironmentOfRadiusN(hmol,radii,atomNum,useHs=True)
    amap = {}
    submol = Chem.PathToSubmol(hmol,env,atomMap=amap)
    atom_smi = Chem.MolToSmiles(submol)
    return('%d'%(hmol.GetAtomWithIdx(atomNum).GetAtomicNum())+atom_smi)

def GetAvals(df):
    num2a = (df.set_index(['A0_NO'])['VALUE'].to_dict())
    list_keys = list(num2a.keys())
    dvals = (np.empty([np.max(list_keys),np.max(list_keys)]))
    for i in range(dvals.shape[0]):
        for j in range(dvals.shape[1]):
            if (i != j) : dvals[i,j] = num2a[i+1]-num2a[j+1]
    dvals[ 0, 5]= 0.0502
    dvals[ 0, 6]= 0.1747
    dvals[ 0, 7]= 0.1671
    dvals[ 5, 6]= 0.0556
    dvals[ 5, 7]= 0.0234
    dvals[ 6, 7]=-0.0346
    # Repitition of the above cooefficients with a negative sign
    dvals[ 5, 0]=-0.0502
    dvals[ 6, 0]=-0.1747
    dvals[ 7, 0]=-0.1671
    dvals[ 6, 5]=-0.0556
    dvals[ 7, 5]=-0.0234
    dvals[ 7, 6]= 0.0346
    return dvals


def GetLogFile(fname,pt_df,rad_df):
    pt_df["symbol"] = pt_df["symbol"].map(str.strip)
    sym2num = pt_df.set_index(['symbol'])['atomicNumber'].to_dict() 
    num2rad = rad_df.set_index(['RAD_NO'])['VALUE'].to_dict()
    xyz_data = []
    charge_data = []
    data = open(fname).readlines()
    id_charges = False
    id_coos = False
    for line in data: 
        if 'CARTESIAN COORDINATES (ANGSTROEM)' in line: 
            id_coos = True 
        elif 'CARTESIAN COORDINATES (A.U.)' in line:
            id_coos = False 
        if 'HIRSHFELD ANALYSIS' in line: 
            id_charges=True
        elif 'TIMINGS' in line: 
            id_charges = False
        if id_charges:charge_data.append(line.strip().split()) 
        if id_coos:xyz_data.append(line.strip().split()) 
    hirCharges = pd.DataFrame(charge_data[7:-4],columns=['N','ATOM','QHir','Spin'])
    hirCharges[['N','QHir','Spin']] = hirCharges[['N','QHir','Spin']].apply(pd.to_numeric)
    hirCharges = hirCharges[['N','QHir']]
    xyzcoos = pd.DataFrame(xyz_data[2:-2],columns=['ATOM','X','Y','Z'])
    xyzcoos[['X','Y','Z']] = xyzcoos[['X','Y','Z']].apply(pd.to_numeric)
    final_data = (pd.concat([xyzcoos,hirCharges],axis=1))
    final_data['AtNum'] = [sym2num[s] for s in final_data.ATOM] 
    final_data['RAD'] = [num2rad[s] for s in final_data.AtNum] 
    return(final_data)

def Distance(a,b):
    return(np.sqrt((a[0]-b[0])**2+(a[1]-b[1])**2+(a[2]-b[2])**2))

def HirshfeldToCM5(df,a0,netcharge): 
    DVALS=GetAvals(a0)
    cm5_charges = []
    alpha = 2.474
    for i,r in df.iterrows(): 
        qcm5 = r.QHir
        for j,p in df.iterrows(): 
            if (r.AtNum != p.AtNum): 
                dist = Distance([r.X,r.Y,r.Z],[p.X,p.Y,p.Z])
                factor = np.exp(-1.0*alpha*(dist-r.RAD-p.RAD))
                qcm5=qcm5+factor*DVALS[r.AtNum-1,p.AtNum-1]
        cm5_charges.append(qcm5)
    df['QCM5']     = np.array(cm5_charges) 
    #df['QCM5']     = df.QCM5 - (netcharge-df.QCM5.sum())/len(df.QCM5) 
    mol = xyz_prep(df)
    df['FPS'] = [AtomFPProgram(mol,atomNum,radii=2) for atomNum in df.index]
    uniq_fps = list(set(df.FPS))
    df['QCM5_AVG'] = [df[df.FPS==i].QCM5.mean() for i in df.FPS]
    df['DEV_CM5'] = df.QCM5_AVG - df.QCM5
    if netcharge == 0:
        df['CM5_final'] = df.QCM5_AVG*1.20
    else: 
        df['CM5_final'] = df.QCM5_AVG*1.00
    #print(df.sum())
    return(df)

def xyz_prep(df):
    opdb = open('inp_orca.xyz', 'w+')
    opdb.write('%3d\n'%(len(df.QCM5)))
    opdb.write('REMARK PolyFlin GENERATED XYZ FILE\n')
    num = 0
    for (i, r) in df.iterrows(): 
        opdb.write('%-6s    %8.3f%8.3f%8.3f\n' %
                   (r.ATOM, r.X, r.Y, r.Z))
    opdb.close()
    
    process = subprocess.Popen('obabel inp_orca.xyz -O UNK.mol', shell=True)
    process.wait()
    
    process = subprocess.Popen('obabel inp_orca.xyz -O inp_orca.pdb', shell=True)
    process.wait()
    
    hmol = Chem.MolFromMolFile('UNK.mol',removeHs=False,sanitize=False)
    return hmol 

def LoadModel(): 
    import json
#    with open("cm5pars.json") as tweetfile:
#        cm5_model = json.loads(tweetfile.read())
    mod_string = '''
{"A0": {"VALUE": {"0": 0.0056, "1": -0.1543, "2": 0.0, "3": 0.0333, "4": -0.103, "5": -0.0446, "6": -0.1072, "7": -0.0802, "8": -0.0629, "9": -0.1088, "10": 0.0184, "11": 0.0, "12": -0.0726, "13": -0.079, "14": -0.0756, "15": -0.0565, "16": -0.0444, "17": -0.0767, "18": 0.013, "19": 0.0, "20": 0.0, "21": 0.0, "22": 0.0, "23": 0.0, "24": 0.0, "25": 0.0, "26": 0.0, "27": 0.0, "28": 0.0, "29": 0.0, "30": -0.0512, "31": -0.0557, "32": -0.0533, "33": -0.0399, "34": -0.0313, "35": -0.0541, "36": 0.0092, "37": 0.0, "38": 0.0, "39": 0.0, "40": 0.0, "41": 0.0, "42": 0.0, "43": 0.0, "44": 0.0, "45": 0.0, "46": 0.0, "47": 0.0, "48": -0.0361, "49": -0.0393, "50": -0.0376, "51": -0.0281, "52": -0.022, "53": -0.0381, "54": 0.0065, "55": 0.0, "56": 0.0, "57": 0.0, "58": 0.0, "59": 0.0, "60": 0.0, "61": 0.0, "62": 0.0, "63": 0.0, "64": 0.0, "65": 0.0, "66": 0.0, "67": 0.0, "68": 0.0, "69": 0.0, "70": 0.0, "71": 0.0, "72": 0.0, "73": 0.0, "74": 0.0, "75": 0.0, "76": 0.0, "77": 0.0, "78": 0.0, "79": 0.0, "80": -0.0255, "81": -0.0277, "82": -0.0265, "83": -0.0198, "84": -0.0155, "85": -0.0269, "86": 0.0046, "87": 0.0, "88": 0.0, "89": 0.0, "90": 0.0, "91": 0.0, "92": 0.0, "93": 0.0, "94": 0.0, "95": 0.0, "96": 0.0, "97": 0.0, "98": 0.0, "99": 0.0, "100": 0.0, "101": 0.0, "102": 0.0, "103": 0.0, "104": 0.0, "105": 0.0, "106": 0.0, "107": 0.0, "108": 0.0, "109": 0.0, "110": 0.0, "111": 0.0, "112": -0.0179, "113": -0.0195, "114": -0.0187, "115": -0.014, "116": -0.011, "117": -0.0189}, "A0_NO": {"0": 1, "1": 2, "2": 3, "3": 4, "4": 5, "5": 6, "6": 7, "7": 8, "8": 9, "9": 10, "10": 11, "11": 12, "12": 13, "13": 14, "14": 15, "15": 16, "16": 17, "17": 18, "18": 19, "19": 20, "20": 21, "21": 22, "22": 23, "23": 24, "24": 25, "25": 26, "26": 27, "27": 28, "28": 29, "29": 30, "30": 31, "31": 32, "32": 33, "33": 34, "34": 35, "35": 36, "36": 37, "37": 38, "38": 39, "39": 40, "40": 41, "41": 42, "42": 43, "43": 44, "44": 45, "45": 46, "46": 47, "47": 48, "48": 49, "49": 50, "50": 51, "51": 52, "52": 53, "53": 54, "54": 55, "55": 56, "56": 57, "57": 58, "58": 59, "59": 60, "60": 61, "61": 62, "62": 63, "63": 64, "64": 65, "65": 66, "66": 67, "67": 68, "68": 69, "69": 70, "70": 71, "71": 72, "72": 73, "73": 74, "74": 75, "75": 76, "76": 77, "77": 78, "78": 79, "79": 80, "80": 81, "81": 82, "82": 83, "83": 84, "84": 85, "85": 86, "86": 87, "87": 88, "88": 89, "89": 90, "90": 91, "91": 92, "92": 93, "93": 94, "94": 95, "95": 96, "96": 97, "97": 98, "98": 99, "99": 100, "100": 101, "101": 102, "102": 103, "103": 104, "104": 105, "105": 106, "106": 107, "107": 108, "108": 109, "109": 110, "110": 111, "111": 112, "112": 113, "113": 114, "114": 115, "115": 116, "116": 117, "117": 118}}, "PeriodicTable": {"yearDiscovered": {"0": "1766", "1": "1868", "2": "1817", "3": "1798", "4": "1807", "5": "Ancient", "6": "1772", "7": "1774", "8": "1670", "9": "1898", "10": "1807", "11": "1808", "12": "Ancient", "13": "1854", "14": "1669", "15": "Ancient", "16": "1774", "17": "1894", "18": "1807", "19": "Ancient", "20": "1876", "21": "1791", "22": "1803", "23": "Ancient", "24": "1774", "25": "Ancient", "26": "Ancient", "27": "1751", "28": "Ancient", "29": "1746", "30": "1875", "31": "1886", "32": "Ancient", "33": "1817", "34": "1826", "35": "1898", "36": "1861", "37": "1790", "38": "1794", "39": "1789", "40": "1801", "41": "1778", "42": "1937", "43": "1827", "44": "1803", "45": "1803", "46": "Ancient", "47": "1817", "48": "1863", "49": "Ancient", "50": "Ancient", "51": "1782", "52": "1811", "53": "1898", "54": "1860", "55": "1808", "56": "1839", "57": "1803", "58": "1885", "59": "1885", "60": "1947", "61": "1853", "62": "1901", "63": "1880", "64": "1843", "65": "1886", "66": "1878", "67": "1842", "68": "1879", "69": "1878", "70": "1907", "71": "1923", "72": "1802", "73": "1783", "74": "1925", "75": "1803", "76": "1803", "77": "Ancient", "78": "Ancient", "79": "Ancient", "80": "1861", "81": "Ancient", "82": "Ancient", "83": "1898", "84": "1940", "85": "1900", "86": "1939", "87": "1898", "88": "1899", "89": "1828", "90": "1913", "91": "1789", "92": "1940", "93": "1940", "94": "1944", "95": "1944", "96": "1949", "97": "1950", "98": "1952", "99": "1952", "100": "1955", "101": "1957", "102": "1961", "103": "1969", "104": "1967", "105": "1974", "106": "1976", "107": "1984", "108": "1982", "109": "1994", "110": "1994", "111": "1996", "112": "2003", "113": "1998", "114": "2003", "115": "2000", "116": "2010", "117": "2002"}, "electronicConfiguration": {"0": "1s1", "1": "1s2", "2": "[He]2s1", "3": "[He]2s2", "4": "[He]2s22p1", "5": "[He]2s22p2", "6": "[He]2s22p3", "7": "[He]2s22p4", "8": "[He]2s22p5", "9": "[He]2s22p6", "10": "[Ne]3s1", "11": "[Ne]3s2", "12": "[Ne]3s23p1", "13": "[Ne]3s23p2", "14": "[Ne]3s23p3", "15": "[Ne]3s23p4", "16": "[Ne]3s23p5", "17": "[Ne]3s23p6", "18": "[Ar]4s1", "19": "[Ar]4s2", "20": "[Ar]3d14s2", "21": "[Ar]3d24s2", "22": "[Ar]3d34s2", "23": "[Ar]3d54s1", "24": "[Ar]3d54s2", "25": "[Ar]3d64s2", "26": "[Ar]3d74s2", "27": "[Ar]3d84s2", "28": "[Ar]3d104s1", "29": "[Ar]3d104s2", "30": "[Ar]3d104s24p1", "31": "[Ar]3d104s24p2", "32": "[Ar]3d104s24p3", "33": "[Ar]3d104s24p4", "34": "[Ar]3d104s24p5", "35": "[Ar]3d104s24p6", "36": "[Kr]5s1", "37": "[Kr]5s2", "38": "[Kr]4d15s2", "39": "[Kr]4d25s2", "40": "[Kr]4d45s1", "41": "[Kr]4d55s1", "42": "[Kr]4d55s2", "43": "[Kr]4d75s1", "44": "[Kr]4d85s1", "45": "[Kr]4d10", "46": "[Kr]4d105s1", "47": "[Kr]4d105s2", "48": "[Kr]4d105s25p1", "49": "[Kr]4d105s25p2", "50": "[Kr]4d105s25p3", "51": "[Kr]4d105s25p4", "52": "[Kr]4d105s25p5", "53": "[Kr]4d105s25p6", "54": "[Xe]6s1", "55": "[Xe]6s2", "56": "[Xe]5d16s2", "57": "[Xe]4f15d16s2", "58": "[Xe]4f36s2", "59": "[Xe]4f46s2", "60": "[Xe]4f56s2", "61": "[Xe]4f66s2", "62": "[Xe]4f76s2", "63": "[Xe]4f75d16s2", "64": "[Xe]4f96s2", "65": "[Xe]4f106s2", "66": "[Xe]4f116s2", "67": "[Xe]4f126s2", "68": "[Xe]4f136s2", "69": "[Xe]4f146s2", "70": "[Xe]4f145d16s2", "71": "[Xe]4f145d26s2", "72": "[Xe]4f145d36s2", "73": "[Xe]4f145d46s2", "74": "[Xe]4f145d56s2", "75": "[Xe]4f145d66s2", "76": "[Xe]4f145d76s2", "77": "[Xe]4f145d96s1", "78": "[Xe]4f145d106s1", "79": "[Xe]4f145d106s2", "80": "[Xe]4f145d106s26p1", "81": "[Xe]4f145d106s26p2", "82": "[Xe]4f145d106s26p3", "83": "[Xe]4f145d106s26p4", "84": "[Xe]4f145d106s26p5", "85": "[Xe]4f145d106s26p6", "86": "[Rn]7s1", "87": "[Rn]7s2", "88": "[Rn]6d17s2", "89": "[Rn]6d27s2", "90": "[Rn]5f26d17s2", "91": "[Rn]5f36d17s2", "92": "[Rn]5f46d17s2", "93": "[Rn]5f67s2", "94": "[Rn]5f77s2", "95": "[Rn]5f76d17s2", "96": "[Rn]5f97s2", "97": "[Rn]5f107s2", "98": "[Rn]5f117s2", "99": "[Rn]5f127s2", "100": "[Rn]5f137s2", "101": "[Rn]5f147s2", "102": "[Rn]5f147s27p1", "103": "[Rn]5f146d27s2", "104": "[Rn]5f146d37s2", "105": "[Rn]5f146d47s2", "106": "[Rn]5f146d57s2", "107": "[Rn]5f146d67s2", "108": "[Rn]5f146d77s2", "109": "[Rn]5f146d97s1", "110": "[Rn]5f146d107s1", "111": "[Rn]5f146d107s2", "112": "[Rn]5f146d107s27p1", "113": "[Rn]5f146d107s27p2", "114": "[Rn]5f146d107s27p3", "115": "[Rn]5f146d107s27p4", "116": "[Rn]5f146d107s27p5", "117": "[Rn]5f146d107s27p6"}, "name": {"0": "Hydrogen", "1": "Helium", "2": "Lithium", "3": "Beryllium", "4": "Boron", "5": "Carbon", "6": "Nitrogen", "7": "Oxygen", "8": "Fluorine", "9": "Neon", "10": "Sodium", "11": "Magnesium", "12": "Aluminum", "13": "Silicon", "14": "Phosphorus", "15": "Sulfur", "16": "Chlorine", "17": "Argon", "18": "Potassium", "19": "Calcium", "20": "Scandium", "21": "Titanium", "22": "Vanadium", "23": "Chromium", "24": "Manganese", "25": "Iron", "26": "Cobalt", "27": "Nickel", "28": "Copper", "29": "Zinc", "30": "Gallium", "31": "Germanium", "32": "Arsenic", "33": "Selenium", "34": "Bromine", "35": "Krypton", "36": "Rubidium", "37": "Strontium", "38": "Yttrium", "39": "Zirconium", "40": "Niobium", "41": "Molybdenum", "42": "Technetium", "43": "Ruthenium", "44": "Rhodium", "45": "Palladium", "46": "Silver", "47": "Cadmium", "48": "Indium", "49": "Tin", "50": "Antimony", "51": "Tellurium", "52": "Iodine", "53": "Xenon", "54": "Cesium", "55": "Barium", "56": "Lanthanum", "57": "Cerium", "58": "Praseodymium", "59": "Neodymium", "60": "Promethium", "61": "Samarium", "62": "Europium", "63": "Gadolinium", "64": "Terbium", "65": "Dysprosium", "66": "Holmium", "67": "Erbium", "68": "Thulium", "69": "Ytterbium", "70": "Lutetium", "71": "Hafnium", "72": "Tantalum", "73": "Tungsten", "74": "Rhenium", "75": "Osmium", "76": "Iridium", "77": "Platinum", "78": "Gold", "79": "Mercury", "80": "Thallium", "81": "Lead", "82": "Bismuth", "83": "Polonium", "84": "Astatine", "85": "Radon", "86": "Francium", "87": "Radium", "88": "Actinium", "89": "Thorium", "90": "Protactinium", "91": "Uranium", "92": "Neptunium", "93": "Plutonium", "94": "Americium", "95": "Curium", "96": "Berkelium", "97": "Californium", "98": "Einsteinium", "99": "Fermium", "100": "Mendelevium", "101": "Nobelium", "102": "Lawrencium", "103": "Rutherfordium", "104": "Dubnium", "105": "Seaborgium", "106": "Bohrium", "107": "Hassium", "108": "Meitnerium", "109": "Darmstadtium", "110": "Roentgenium", "111": "Copernicium", "112": "Nihonium", "113": "Flerovium", "114": "Moscovium", "115": "Livermorium", "116": "Tennessine", "117": "Oganesson"}, "groupBlock": {"0": "nonmetal", "1": "noblegas", "2": "alkalimetal", "3": "alkalineearthmetal", "4": "metalloid", "5": "nonmetal", "6": "nonmetal", "7": "nonmetal", "8": "halogen", "9": "noblegas", "10": "alkalimetal", "11": "alkalineearthmetal", "12": "metal", "13": "metalloid", "14": "nonmetal", "15": "nonmetal", "16": "halogen", "17": "noblegas", "18": "alkalimetal", "19": "alkalineearthmetal", "20": "transitionmetal", "21": "transitionmetal", "22": "transitionmetal", "23": "transitionmetal", "24": "transitionmetal", "25": "transitionmetal", "26": "transitionmetal", "27": "transitionmetal", "28": "transitionmetal", "29": "transitionmetal", "30": "metal", "31": "metalloid", "32": "metalloid", "33": "nonmetal", "34": "halogen", "35": "noblegas", "36": "alkalimetal", "37": "alkalineearthmetal", "38": "transitionmetal", "39": "transitionmetal", "40": "transitionmetal", "41": "transitionmetal", "42": "transitionmetal", "43": "transitionmetal", "44": "transitionmetal", "45": "transitionmetal", "46": "transitionmetal", "47": "transitionmetal", "48": "metal", "49": "metal", "50": "metalloid", "51": "metalloid", "52": "halogen", "53": "noblegas", "54": "alkalimetal", "55": "alkalineearthmetal", "56": "lanthanoid", "57": "lanthanoid", "58": "lanthanoid", "59": "lanthanoid", "60": "lanthanoid", "61": "lanthanoid", "62": "lanthanoid", "63": "lanthanoid", "64": "lanthanoid", "65": "lanthanoid", "66": "lanthanoid", "67": "lanthanoid", "68": "lanthanoid", "69": "lanthanoid", "70": "lanthanoid", "71": "transitionmetal", "72": "transitionmetal", "73": "transitionmetal", "74": "transitionmetal", "75": "transitionmetal", "76": "transitionmetal", "77": "transitionmetal", "78": "transitionmetal", "79": "transitionmetal", "80": "metal", "81": "metal", "82": "metal", "83": "metalloid", "84": "halogen", "85": "noblegas", "86": "alkalimetal", "87": "alkalineearthmetal", "88": "actinoid", "89": "actinoid", "90": "actinoid", "91": "actinoid", "92": "actinoid", "93": "actinoid", "94": "actinoid", "95": "actinoid", "96": "actinoid", "97": "actinoid", "98": "actinoid", "99": "actinoid", "100": "actinoid", "101": "actinoid", "102": "transitionmetal", "103": "transitionmetal", "104": "transitionmetal", "105": "transitionmetal", "106": "transitionmetal", "107": "transitionmetal", "108": "transitionmetal", "109": "transitionmetal", "110": "transitionmetal", "111": "transitionmetal", "112": "post-transitionmetal", "113": "post-transitionmetal", "114": "post-transitionmetal", "115": "post-transitionmetal", "116": "post-transitionmetal", "117": "noblegas"}, "standardState": {"0": "gas", "1": "gas", "2": "solid", "3": "solid", "4": "solid", "5": "solid", "6": "gas", "7": "gas", "8": "gas", "9": "gas", "10": "solid", "11": "solid", "12": "solid", "13": "solid", "14": "solid", "15": "solid", "16": "gas", "17": "gas", "18": "solid", "19": "solid", "20": "solid", "21": "solid", "22": "solid", "23": "solid", "24": "solid", "25": "solid", "26": "solid", "27": "solid", "28": "solid", "29": "solid", "30": "solid", "31": "solid", "32": "solid", "33": "solid", "34": "liquid", "35": "gas", "36": "solid", "37": "solid", "38": "solid", "39": "solid", "40": "solid", "41": "solid", "42": "solid", "43": "solid", "44": "solid", "45": "solid", "46": "solid", "47": "solid", "48": "solid", "49": "solid", "50": "solid", "51": "solid", "52": "solid", "53": "gas", "54": "solid", "55": "solid", "56": "solid", "57": "solid", "58": "solid", "59": "solid", "60": "solid", "61": "solid", "62": "solid", "63": "solid", "64": "solid", "65": "solid", "66": "solid", "67": "solid", "68": "solid", "69": "solid", "70": "solid", "71": "solid", "72": "solid", "73": "solid", "74": "solid", "75": "solid", "76": "solid", "77": "solid", "78": "solid", "79": "liquid", "80": "solid", "81": "solid", "82": "solid", "83": "solid", "84": "solid", "85": "gas", "86": "solid", "87": "solid", "88": "solid", "89": "solid", "90": "solid", "91": "solid", "92": "solid", "93": "solid", "94": "solid", "95": "solid", "96": "solid", "97": "solid", "98": "solid", "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "atomicRadius": {"0": 37.0, "1": 32.0, "2": 134.0, "3": 90.0, "4": 82.0, "5": 77.0, "6": 75.0, "7": 73.0, "8": 71.0, "9": 69.0, "10": 154.0, "11": 130.0, "12": 118.0, "13": 111.0, "14": 106.0, "15": 102.0, "16": 99.0, "17": 97.0, "18": 196.0, "19": 174.0, "20": 144.0, "21": 136.0, "22": 125.0, "23": 127.0, "24": 139.0, "25": 125.0, "26": 126.0, "27": 121.0, "28": 138.0, "29": 131.0, "30": 126.0, "31": 122.0, "32": 119.0, "33": 116.0, "34": 114.0, "35": 110.0, "36": 211.0, "37": 192.0, "38": 162.0, "39": 148.0, "40": 137.0, "41": 145.0, "42": 156.0, "43": 126.0, "44": 135.0, "45": 131.0, "46": 153.0, "47": 148.0, "48": 144.0, "49": 141.0, "50": 138.0, "51": 135.0, "52": 133.0, "53": 130.0, "54": 225.0, "55": 198.0, "56": 169.0, "57": NaN, "58": NaN, "59": NaN, "60": NaN, "61": NaN, "62": NaN, "63": NaN, "64": NaN, "65": NaN, "66": NaN, "67": NaN, "68": NaN, "69": NaN, "70": 160.0, "71": 150.0, "72": 138.0, "73": 146.0, "74": 159.0, "75": 128.0, "76": 137.0, "77": 128.0, "78": 144.0, "79": 149.0, "80": 148.0, "81": 147.0, "82": 146.0, "83": NaN, "84": NaN, "85": 145.0, "86": NaN, "87": NaN, "88": NaN, "89": NaN, "90": NaN, "91": NaN, "92": NaN, "93": NaN, "94": NaN, "95": NaN, "96": NaN, "97": NaN, "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "oxidationStates": {"0": "-1,1", "1": NaN, "2": "1", "3": "2", "4": "1,2,3", "5": "-4,-3,-2,-1,1,2,3,4", "6": "-3,-2,-1,1,2,3,4,5", "7": "-2,-1,1,2", "8": "-1", "9": NaN, "10": "-1,1", "11": "1,2", "12": "1,3", "13": "-4,-3,-2,-1,1,2,3,4", "14": "-3,-2,-1,1,2,3,4,5", "15": "-2,-1,1,2,3,4,5,6", "16": "-1,1,2,3,4,5,6,7", "17": NaN, "18": "1", "19": "2", "20": "1,2,3", "21": "-1,2,3,4", "22": "-1,2,3,4", "23": "-2,-1,1,2,3,4,5,6", "24": "-3,-2,-1,1,2,3,4,5,6,7", "25": "-2,-1,1,2,3,4,5,6", "26": "-1,1,2,3,4,5", "27": "-1,1,2,3,4", "28": "1,2,3,4", "29": "2", "30": "1,2,3", "31": "-4,1,2,3,4", "32": "-3,2,3,5", "33": "-2,2,4,6", "34": "-1,1,3,4,5,7", "35": "2", "36": "1", "37": "2", "38": "1,2,3", "39": "1,2,3,4", "40": "-1,2,3,4,5", "41": "-2,-1,1,2,3,4,5,6", "42": "-3,-1,1,2,3,4,5,6,7", "43": "-2,1,2,3,4,5,6,7,8", "44": "-1,1,2,3,4,5,6", "45": "2,4", "46": "1,2,3", "47": "2", "48": "1,2,3", "49": "-4,2,4", "50": "-3,3,5", "51": "-2,2,4,5,6", "52": "-1,1,3,5,7", "53": "2,4,6,8", "54": "1", "55": "2", "56": "2,3", "57": "2,3,4", "58": "2,3,4", "59": "2,3", "60": "3", "61": "2,3", "62": "2,3", "63": "1,2,3", "64": "1,3,4", "65": "2,3", "66": "3", "67": "3", "68": "2,3", "69": "2,3", "70": "3", "71": "2,3,4", "72": "-1,2,3,4,5", "73": "-2,-1,1,2,3,4,5,6", "74": "-3,-1,1,2,3,4,5,6,7", "75": "-2,-1,1,2,3,4,5,6,7,8", "76": "-3,-1,1,2,3,4,5,6", "77": "2,4,5,6", "78": "-1,1,2,3,5", "79": "1,2,4", "80": "1,3", "81": "-4,2,4", "82": "-3,3,5", "83": "-2,2,4,6", "84": "-1,1,3,5", "85": "2", "86": "1", "87": "2", "88": "3", "89": "2,3,4", "90": "3,4,5", "91": "3,4,5,6", "92": "3,4,5,6,7", "93": "3,4,5,6,7", "94": "2,3,4,5,6", "95": "3,4", "96": "3,4", "97": "2,3,4", "98": "2,3", "99": "2,3", "100": "2,3", "101": "2,3", "102": "3", "103": "4", "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "atomicMass": {"0": "1.00794(4)", "1": "4.002602(2)", "2": "6.941(2)", "3": "9.012182(3)", "4": "10.811(7)", "5": "12.0107(8)", "6": "14.0067(2)", "7": "15.9994(3)", "8": "18.9984032(5)", "9": "20.1797(6)", "10": "22.98976928(2)", "11": "24.3050(6)", "12": "26.9815386(8)", "13": "28.0855(3)", "14": "30.973762(2)", "15": "32.065(5)", "16": "35.453(2)", "17": "39.948(1)", "18": "39.0983(1)", "19": "40.078(4)", "20": "44.955912(6)", "21": "47.867(1)", "22": "50.9415(1)", "23": "51.9961(6)", "24": "54.938045(5)", "25": "55.845(2)", "26": "58.933195(5)", "27": "58.6934(4)", "28": "63.546(3)", "29": "65.38(2)", "30": "69.723(1)", "31": "72.64(1)", "32": "74.92160(2)", "33": "78.96(3)", "34": "79.904(1)", "35": "83.798(2)", "36": "85.4678(3)", "37": "87.62(1)", "38": "88.90585(2)", "39": "91.224(2)", "40": "92.90638(2)", "41": "95.96(2)", "42": "[98]", "43": "101.07(2)", "44": "102.90550(2)", "45": "106.42(1)", "46": "107.8682(2)", "47": "112.411(8)", "48": "114.818(3)", "49": "118.710(7)", "50": "121.760(1)", "51": "127.60(3)", "52": "126.90447(3)", "53": "131.293(6)", "54": "132.9054519(2)", "55": "137.327(7)", "56": "138.90547(7)", "57": "140.116(1)", "58": "140.90765(2)", "59": "144.242(3)", "60": "[145]", "61": "150.36(2)", "62": "151.964(1)", "63": "157.25(3)", "64": "158.92535(2)", "65": "162.500(1)", "66": "164.93032(2)", "67": "167.259(3)", "68": "168.93421(2)", "69": "173.054(5)", "70": "174.9668(1)", "71": "178.49(2)", "72": "180.94788(2)", "73": "183.84(1)", "74": "186.207(1)", "75": "190.23(3)", "76": "192.217(3)", "77": "195.084(9)", "78": "196.966569(4)", "79": "200.59(2)", "80": "204.3833(2)", "81": "207.2(1)", "82": "208.98040(1)", "83": "[209]", "84": "[210]", "85": "[222]", "86": "[223]", "87": "[226]", "88": "[227]", "89": "232.03806(2)", "90": "231.03588(2)", "91": "238.02891(3)", "92": "[237]", "93": "[244]", "94": "[243]", "95": "[247]", "96": "[247]", "97": "[251]", "98": "[252]", "99": "[257]", "100": "[258]", "101": "[259]", "102": "[262]", "103": "[267]", "104": "[268]", "105": "[271]", "106": "[272]", "107": "[270]", "108": "[276]", "109": "[281]", "110": "[280]", "111": "[285]", "112": "[284]", "113": "[289]", "114": "[288]", "115": "[293]", "116": "[294]", "117": "[294]"}, "cpkHexColor": {"0": "FFFFFF", "1": "D9FFFF", "2": "CC80FF", "3": "C2FF00", "4": "FFB5B5", "5": "909090", "6": "3050F8", "7": "FF0D0D", "8": "90E050", "9": "B3E3F5", "10": "AB5CF2", "11": "8AFF00", "12": "BFA6A6", "13": "F0C8A0", "14": "FF8000", "15": "FFFF30", "16": "1FF01F", "17": "80D1E3", "18": "8F40D4", "19": "3DFF00", "20": "E6E6E6", "21": "BFC2C7", "22": "A6A6AB", "23": "8A99C7", "24": "9C7AC7", "25": "E06633", "26": "F090A0", "27": "50D050", "28": "C88033", "29": "7D80B0", "30": "C28F8F", "31": "668F8F", "32": "BD80E3", "33": "FFA100", "34": "A62929", "35": "5CB8D1", "36": "702EB0", "37": "00FF00", "38": "94FFFF", "39": "94E0E0", "40": "73C2C9", "41": "54B5B5", "42": "3B9E9E", "43": "248F8F", "44": "0A7D8C", "45": "6985", "46": "C0C0C0", "47": "FFD98F", "48": "A67573", "49": "668080", "50": "9E63B5", "51": "D47A00", "52": "940094", "53": "429EB0", "54": "57178F", "55": "00C900", "56": "70D4FF", "57": "FFFFC7", "58": "D9FFC7", "59": "C7FFC7", "60": "A3FFC7", "61": "8FFFC7", "62": "61FFC7", "63": "45FFC7", "64": "30FFC7", "65": "1FFFC7", "66": "00FF9C", "67": "0.00E+00", "68": "00D452", "69": "00BF38", "70": "00AB24", "71": "4DC2FF", "72": "4DA6FF", "73": "2194D6", "74": "267DAB", "75": "266696", "76": "175487", "77": "D0D0E0", "78": "FFD123", "79": "B8B8D0", "80": "A6544D", "81": "575961", "82": "9E4FB5", "83": "AB5C00", "84": "754F45", "85": "428296", "86": "420066", "87": "007D00", "88": "70ABFA", "89": "00BAFF", "90": "00A1FF", "91": "008FFF", "92": "0080FF", "93": "006BFF", "94": "545CF2", "95": "785CE3", "96": "8A4FE3", "97": "A136D4", "98": "B31FD4", "99": "B31FBA", "100": "B30DA6", "101": "BD0D87", "102": "C70066", "103": "CC0059", "104": "D1004F", "105": "D90045", "106": "E00038", "107": "E6002E", "108": "EB0026", "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "density": {"0": 8.99e-05, "1": 0.0001785, "2": 0.535, "3": 1.848, "4": 2.46, "5": 2.26, "6": 0.001251, "7": 0.001429, "8": 0.001696, "9": 0.0009, "10": 0.968, "11": 1.7380000000000002, "12": 2.7, "13": 2.33, "14": 1.8230000000000002, "15": 1.96, "16": 0.003214, "17": 0.001784, "18": 0.856, "19": 1.55, "20": 2.985, "21": 4.507, "22": 6.11, "23": 7.14, "24": 7.47, "25": 7.874, "26": 8.9, "27": 8.908, "28": 8.92, "29": 7.14, "30": 5.904, "31": 5.3229999999999995, "32": 5.727, "33": 4.819, "34": 3.12, "35": 0.00375, "36": 1.5319999999999998, "37": 2.63, "38": 4.4719999999999995, "39": 6.511, "40": 8.57, "41": 10.28, "42": 11.5, "43": 12.37, "44": 12.45, "45": 12.023, "46": 10.49, "47": 8.65, "48": 7.31, "49": 7.31, "50": 6.697, "51": 6.24, "52": 4.94, "53": 0.0059, "54": 1.879, "55": 3.51, "56": 6.146, "57": 6.689, "58": 6.64, "59": 7.01, "60": 7.263999999999999, "61": 7.353, "62": 5.244, "63": 7.901, "64": 8.219, "65": 8.551, "66": 8.795, "67": 9.066, "68": 9.321, "69": 6.57, "70": 9.841000000000001, "71": 13.31, "72": 16.65, "73": 19.25, "74": 21.02, "75": 22.61, "76": 22.65, "77": 21.09, "78": 19.3, "79": 13.534, "80": 11.85, "81": 11.34, "82": 9.78, "83": 9.196, "84": NaN, "85": 0.009729999999999999, "86": NaN, "87": 5.0, "88": 10.07, "89": 11.724, "90": 15.37, "91": 19.05, "92": 20.45, "93": 19.816, "94": NaN, "95": 13.51, "96": 14.78, "97": 15.1, "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "ionRadius": {"0": NaN, "1": NaN, "2": "76(+1)", "3": "45(+2)", "4": "27(+3)", "5": "16(+4)", "6": "146(-3)", "7": "140(-2)", "8": "133(-1)", "9": NaN, "10": "102(+1)", "11": "72(+2)", "12": "53.5(+3)", "13": "40(+4)", "14": "44(+3)", "15": "184(-2)", "16": "181(-1)", "17": NaN, "18": "138(+1)", "19": "100(+2)", "20": "74.5(+3)", "21": "86(+2)", "22": "79(+2)", "23": "80(+2*)", "24": "67(+2)", "25": "78(+2*)", "26": "74.5(+2*)", "27": "69(+2)", "28": "77(+1)", "29": "74(+2)", "30": "62(+3)", "31": "73(+2)", "32": "58(+3)", "33": "198(-2)", "34": "196(-1)", "35": NaN, "36": "152(+1)", "37": "118(+2)", "38": "90(+3)", "39": "72(+4)", "40": "72(+3)", "41": "69(+3)", "42": "64.5(+4)", "43": "68(+3)", "44": "66.5(+3)", "45": "59(+1)", "46": "115(+1)", "47": "95(+2)", "48": "80(+3)", "49": "112(+2)", "50": "76(+3)", "51": "221(-2)", "52": "220(-1)", "53": "48(+8)", "54": "167(+1)", "55": "135(+2)", "56": "103.2(+3)", "57": "102(+3)", "58": "99(+3)", "59": "129(+2)", "60": "97(+3)", "61": "122(+2)", "62": "117(+2)", "63": "93.8(+3)", "64": "92.3(+3)", "65": "107(+2)", "66": "90.1(+3)", "67": "89(+3)", "68": "103(+2)", "69": "102(+2)", "70": "86.1(+3)", "71": "71(+4)", "72": "72(+3)", "73": "66(+4)", "74": "63(+4)", "75": "63(+4)", "76": "68(+3)", "77": "86(+2)", "78": "137(+1)", "79": "119(+1)", "80": "150(+1)", "81": "119(+2)", "82": "103(+3)", "83": "94(+4)", "84": "62(+7)", "85": NaN, "86": "180(+1)", "87": "148(+2)", "88": "112(+3)", "89": "94(+4)", "90": "104(+3)", "91": "102.5(+3)", "92": "110(+2)", "93": "100(+3)", "94": "126(+2)", "95": "97(+3)", "96": "96(+3)", "97": "95(+3)", "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "electronegativity": {"0": 2.2, "1": NaN, "2": 0.98, "3": 1.57, "4": 2.04, "5": 2.55, "6": 3.04, "7": 3.44, "8": 3.98, "9": NaN, "10": 0.93, "11": 1.31, "12": 1.61, "13": 1.9, "14": 2.19, "15": 2.58, "16": 3.16, "17": NaN, "18": 0.82, "19": 1.0, "20": 1.36, "21": 1.54, "22": 1.63, "23": 1.66, "24": 1.55, "25": 1.83, "26": 1.88, "27": 1.91, "28": 1.9, "29": 1.65, "30": 1.81, "31": 2.01, "32": 2.18, "33": 2.55, "34": 2.96, "35": NaN, "36": 0.82, "37": 0.95, "38": 1.22, "39": 1.33, "40": 1.6, "41": 2.16, "42": 1.9, "43": 2.2, "44": 2.28, "45": 2.2, "46": 1.93, "47": 1.69, "48": 1.78, "49": 1.96, "50": 2.05, "51": 2.1, "52": 2.66, "53": NaN, "54": 0.79, "55": 0.89, "56": 1.1, "57": 1.12, "58": 1.13, "59": 1.14, "60": 1.13, "61": 1.17, "62": 1.2, "63": 1.2, "64": 1.2, "65": 1.22, "66": 1.23, "67": 1.24, "68": 1.25, "69": 1.1, "70": 1.27, "71": 1.3, "72": 1.5, "73": 2.36, "74": 1.9, "75": 2.2, "76": 2.2, "77": 2.28, "78": 2.54, "79": 2.0, "80": 2.04, "81": 2.33, "82": 2.02, "83": 2.0, "84": 2.2, "85": NaN, "86": 0.7, "87": 0.9, "88": 1.1, "89": 1.3, "90": 1.5, "91": 1.38, "92": 1.36, "93": 1.28, "94": 1.3, "95": 1.3, "96": 1.3, "97": 1.3, "98": 1.3, "99": 1.3, "100": 1.3, "101": 1.3, "102": 1.3, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "atomicNumber": {"0": 1, "1": 2, "2": 3, "3": 4, "4": 5, "5": 6, "6": 7, "7": 8, "8": 9, "9": 10, "10": 11, "11": 12, "12": 13, "13": 14, "14": 15, "15": 16, "16": 17, "17": 18, "18": 19, "19": 20, "20": 21, "21": 22, "22": 23, "23": 24, "24": 25, "25": 26, "26": 27, "27": 28, "28": 29, "29": 30, "30": 31, "31": 32, "32": 33, "33": 34, "34": 35, "35": 36, "36": 37, "37": 38, "38": 39, "39": 40, "40": 41, "41": 42, "42": 43, "43": 44, "44": 45, "45": 46, "46": 47, "47": 48, "48": 49, "49": 50, "50": 51, "51": 52, "52": 53, "53": 54, "54": 55, "55": 56, "56": 57, "57": 58, "58": 59, "59": 60, "60": 61, "61": 62, "62": 63, "63": 64, "64": 65, "65": 66, "66": 67, "67": 68, "68": 69, "69": 70, "70": 71, "71": 72, "72": 73, "73": 74, "74": 75, "75": 76, "76": 77, "77": 78, "78": 79, "79": 80, "80": 81, "81": 82, "82": 83, "83": 84, "84": 85, "85": 86, "86": 87, "87": 88, "88": 89, "89": 90, "90": 91, "91": 92, "92": 93, "93": 94, "94": 95, "95": 96, "96": 97, "97": 98, "98": 99, "99": 100, "100": 101, "101": 102, "102": 103, "103": 104, "104": 105, "105": 106, "106": 107, "107": 108, "108": 109, "109": 110, "110": 111, "111": 112, "112": 113, "113": 114, "114": 115, "115": 116, "116": 117, "117": 118}, "symbol": {"0": "H", "1": "He", "2": "Li", "3": "Be", "4": "B", "5": "C", "6": "N", "7": "O", "8": "F", "9": "Ne", "10": "Na", "11": "Mg", "12": "Al", "13": "Si", "14": "P", "15": "S", "16": "Cl", "17": "Ar", "18": "K", "19": "Ca", "20": "Sc", "21": "Ti", "22": "V", "23": "Cr", "24": "Mn", "25": "Fe", "26": "Co", "27": "Ni", "28": "Cu", "29": "Zn", "30": "Ga", "31": "Ge", "32": "As", "33": "Se", "34": "Br", "35": "Kr", "36": "Rb", "37": "Sr", "38": "Y", "39": "Zr", "40": "Nb", "41": "Mo", "42": "Tc", "43": "Ru", "44": "Rh", "45": "Pd", "46": "Ag", "47": "Cd", "48": "In", "49": "Sn", "50": "Sb", "51": "Te", "52": "I", "53": "Xe", "54": "Cs", "55": "Ba", "56": "La", "57": "Ce", "58": "Pr", "59": "Nd", "60": "Pm", "61": "Sm", "62": "Eu", "63": "Gd", "64": "Tb", "65": "Dy", "66": "Ho", "67": "Er", "68": "Tm", "69": "Yb", "70": "Lu", "71": "Hf", "72": "Ta", "73": "W", "74": "Re", "75": "Os", "76": "Ir", "77": "Pt", "78": "Au", "79": "Hg", "80": "Tl", "81": "Pb", "82": "Bi", "83": "Po", "84": "At", "85": "Rn", "86": "Fr", "87": "Ra", "88": "Ac", "89": "Th", "90": "Pa", "91": "U", "92": "Np", "93": "Pu", "94": "Am", "95": "Cm", "96": "Bk", "97": "Cf", "98": "Es", "99": "Fm", "100": "Md", "101": "No", "102": "Lr", "103": "Rf", "104": "Db", "105": "Sg", "106": "Bh", "107": "Hs", "108": "Mt", "109": "Ds", "110": "Rg", "111": "Cn", "112": "Nh", "113": "Fl", "114": "Mc", "115": "Lv", "116": "Ts", "117": "Og"}, "bondingType": {"0": "diatomic", "1": "atomic", "2": "metallic", "3": "metallic", "4": "covalentnetwork", "5": "covalentnetwork", "6": "diatomic", "7": "diatomic", "8": "atomic", "9": "atomic", "10": "metallic", "11": "metallic", "12": "metallic", "13": "metallic", "14": "covalentnetwork", "15": "covalentnetwork", "16": "covalentnetwork", "17": "atomic", "18": "metallic", "19": "metallic", "20": "metallic", "21": "metallic", "22": "metallic", "23": "metallic", "24": "metallic", "25": "metallic", "26": "metallic", "27": "metallic", "28": "metallic", "29": "metallic", "30": "metallic", "31": "metallic", "32": "metallic", "33": "metallic", "34": "covalentnetwork", "35": "atomic", "36": "metallic", "37": "metallic", "38": "metallic", "39": "metallic", "40": "metallic", "41": "metallic", "42": "metallic", "43": "metallic", "44": "metallic", "45": "metallic", "46": "metallic", "47": "metallic", "48": "metallic", "49": "metallic", "50": "metallic", "51": "metallic", "52": "covalentnetwork", "53": "atomic", "54": "metallic", "55": "metallic", "56": "metallic", "57": "metallic", "58": "metallic", "59": "metallic", "60": "metallic", "61": "metallic", "62": "metallic", "63": "metallic", "64": "metallic", "65": "metallic", "66": "metallic", "67": "metallic", "68": "metallic", "69": "metallic", "70": "metallic", "71": "metallic", "72": "metallic", "73": "metallic", "74": "metallic", "75": "metallic", "76": "metallic", "77": "metallic", "78": "metallic", "79": "metallic", "80": "metallic", "81": "metallic", "82": "metallic", "83": "metallic", "84": "covalentnetwork", "85": "atomic", "86": "metallic", "87": "metallic", "88": "metallic", "89": "metallic", "90": "metallic", "91": "metallic", "92": "metallic", "93": "metallic", "94": "metallic", "95": "metallic", "96": "metallic", "97": "metallic", "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "electronAffinity": {"0": -73.0, "1": 0.0, "2": -60.0, "3": 0.0, "4": -27.0, "5": -154.0, "6": -7.0, "7": -141.0, "8": -328.0, "9": 0.0, "10": -53.0, "11": 0.0, "12": -43.0, "13": -134.0, "14": -72.0, "15": -200.0, "16": -349.0, "17": 0.0, "18": -48.0, "19": -2.0, "20": -18.0, "21": -8.0, "22": -51.0, "23": -64.0, "24": 0.0, "25": -16.0, "26": -64.0, "27": -112.0, "28": -118.0, "29": 0.0, "30": -29.0, "31": -119.0, "32": -78.0, "33": -195.0, "34": -325.0, "35": 0.0, "36": -47.0, "37": -5.0, "38": -30.0, "39": -41.0, "40": -86.0, "41": -72.0, "42": -53.0, "43": -101.0, "44": -110.0, "45": -54.0, "46": -126.0, "47": 0.0, "48": -29.0, "49": -107.0, "50": -103.0, "51": -190.0, "52": -295.0, "53": 0.0, "54": -46.0, "55": -14.0, "56": -48.0, "57": -50.0, "58": -50.0, "59": -50.0, "60": -50.0, "61": -50.0, "62": -50.0, "63": -50.0, "64": -50.0, "65": -50.0, "66": -50.0, "67": -50.0, "68": -50.0, "69": -50.0, "70": -50.0, "71": 0.0, "72": -31.0, "73": -79.0, "74": -15.0, "75": -106.0, "76": -151.0, "77": -205.0, "78": -223.0, "79": 0.0, "80": -19.0, "81": -35.0, "82": -91.0, "83": -183.0, "84": -270.0, "85": NaN, "86": NaN, "87": NaN, "88": NaN, "89": NaN, "90": NaN, "91": NaN, "92": NaN, "93": NaN, "94": NaN, "95": NaN, "96": NaN, "97": NaN, "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "ionizationEnergy": {"0": 1312.0, "1": 2372.0, "2": 520.0, "3": 900.0, "4": 801.0, "5": 1087.0, "6": 1402.0, "7": 1314.0, "8": 1681.0, "9": 2081.0, "10": 496.0, "11": 738.0, "12": 578.0, "13": 787.0, "14": 1012.0, "15": 1000.0, "16": 1251.0, "17": 1521.0, "18": 419.0, "19": 590.0, "20": 633.0, "21": 659.0, "22": 651.0, "23": 653.0, "24": 717.0, "25": 763.0, "26": 760.0, "27": 737.0, "28": 746.0, "29": 906.0, "30": 579.0, "31": 762.0, "32": 947.0, "33": 941.0, "34": 1140.0, "35": 1351.0, "36": 403.0, "37": 550.0, "38": 600.0, "39": 640.0, "40": 652.0, "41": 684.0, "42": 702.0, "43": 710.0, "44": 720.0, "45": 804.0, "46": 731.0, "47": 868.0, "48": 558.0, "49": 709.0, "50": 834.0, "51": 869.0, "52": 1008.0, "53": 1170.0, "54": 376.0, "55": 503.0, "56": 538.0, "57": 534.0, "58": 527.0, "59": 533.0, "60": 540.0, "61": 545.0, "62": 547.0, "63": 593.0, "64": 566.0, "65": 573.0, "66": 581.0, "67": 589.0, "68": 597.0, "69": 603.0, "70": 524.0, "71": 659.0, "72": 761.0, "73": 770.0, "74": 760.0, "75": 840.0, "76": 880.0, "77": 870.0, "78": 890.0, "79": 1007.0, "80": 589.0, "81": 716.0, "82": 703.0, "83": 812.0, "84": 920.0, "85": 1037.0, "86": 380.0, "87": 509.0, "88": 499.0, "89": 587.0, "90": 568.0, "91": 598.0, "92": 605.0, "93": 585.0, "94": 578.0, "95": 581.0, "96": 601.0, "97": 608.0, "98": 619.0, "99": 627.0, "100": 635.0, "101": 642.0, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "vanDelWaalsRadius": {"0": 120.0, "1": 140.0, "2": 182.0, "3": NaN, "4": NaN, "5": 170.0, "6": 155.0, "7": 152.0, "8": 147.0, "9": 154.0, "10": 227.0, "11": 173.0, "12": NaN, "13": 210.0, "14": 180.0, "15": 180.0, "16": 175.0, "17": 188.0, "18": 275.0, "19": NaN, "20": NaN, "21": NaN, "22": NaN, "23": NaN, "24": NaN, "25": NaN, "26": NaN, "27": 163.0, "28": 140.0, "29": 139.0, "30": 187.0, "31": NaN, "32": 185.0, "33": 190.0, "34": 185.0, "35": 202.0, "36": NaN, "37": NaN, "38": NaN, "39": NaN, "40": NaN, "41": NaN, "42": NaN, "43": NaN, "44": NaN, "45": 163.0, "46": 172.0, "47": 158.0, "48": 193.0, "49": 217.0, "50": NaN, "51": 206.0, "52": 198.0, "53": 216.0, "54": NaN, "55": NaN, "56": NaN, "57": NaN, "58": NaN, "59": NaN, "60": NaN, "61": NaN, "62": NaN, "63": NaN, "64": NaN, "65": NaN, "66": NaN, "67": NaN, "68": NaN, "69": NaN, "70": NaN, "71": NaN, "72": NaN, "73": NaN, "74": NaN, "75": NaN, "76": NaN, "77": 175.0, "78": 166.0, "79": 155.0, "80": 196.0, "81": 202.0, "82": NaN, "83": NaN, "84": NaN, "85": NaN, "86": NaN, "87": NaN, "88": NaN, "89": NaN, "90": NaN, "91": 186.0, "92": NaN, "93": NaN, "94": NaN, "95": NaN, "96": NaN, "97": NaN, "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "boilingPoint": {"0": 20.0, "1": 4.0, "2": 1615.0, "3": 2743.0, "4": 4273.0, "5": 4300.0, "6": 77.0, "7": 90.0, "8": 85.0, "9": 27.0, "10": 1156.0, "11": 1363.0, "12": 2792.0, "13": 3173.0, "14": 554.0, "15": 718.0, "16": 239.0, "17": 87.0, "18": 1032.0, "19": 1757.0, "20": 3103.0, "21": 3560.0, "22": 3680.0, "23": 2944.0, "24": 2334.0, "25": 3134.0, "26": 3200.0, "27": 3186.0, "28": 3200.0, "29": 1180.0, "30": 2477.0, "31": 3093.0, "32": 887.0, "33": 958.0, "34": 332.0, "35": 120.0, "36": 961.0, "37": 1655.0, "38": 3618.0, "39": 4682.0, "40": 5017.0, "41": 4912.0, "42": 4538.0, "43": 4423.0, "44": 3968.0, "45": 3236.0, "46": 2435.0, "47": 1040.0, "48": 2345.0, "49": 2875.0, "50": 1860.0, "51": 1261.0, "52": 457.0, "53": 165.0, "54": 944.0, "55": 2143.0, "56": 3737.0, "57": 3633.0, "58": 3563.0, "59": 3373.0, "60": 3273.0, "61": 2076.0, "62": 1800.0, "63": 3523.0, "64": 3503.0, "65": 2840.0, "66": 2973.0, "67": 3141.0, "68": 2223.0, "69": 1469.0, "70": 3675.0, "71": 4876.0, "72": 5731.0, "73": 5828.0, "74": 5869.0, "75": 5285.0, "76": 4701.0, "77": 4098.0, "78": 3129.0, "79": 630.0, "80": 1746.0, "81": 2022.0, "82": 1837.0, "83": 1235.0, "84": NaN, "85": 211.0, "86": NaN, "87": 2010.0, "88": 3473.0, "89": 5093.0, "90": 4273.0, "91": 4200.0, "92": 4273.0, "93": 3503.0, "94": 2284.0, "95": 3383.0, "96": NaN, "97": NaN, "98": NaN, "99": NaN, "100": NaN, "101": NaN, "102": NaN, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}, "meltingPoint": {"0": 14.0, "1": NaN, "2": 454.0, "3": 1560.0, "4": 2348.0, "5": 3823.0, "6": 63.0, "7": 55.0, "8": 54.0, "9": 25.0, "10": 371.0, "11": 923.0, "12": 933.0, "13": 1687.0, "14": 317.0, "15": 388.0, "16": 172.0, "17": 84.0, "18": 337.0, "19": 1115.0, "20": 1814.0, "21": 1941.0, "22": 2183.0, "23": 2180.0, "24": 1519.0, "25": 1811.0, "26": 1768.0, "27": 1728.0, "28": 1358.0, "29": 693.0, "30": 303.0, "31": 1211.0, "32": 1090.0, "33": 494.0, "34": 266.0, "35": 116.0, "36": 312.0, "37": 1050.0, "38": 1799.0, "39": 2128.0, "40": 2750.0, "41": 2896.0, "42": 2430.0, "43": 2607.0, "44": 2237.0, "45": 1828.0, "46": 1235.0, "47": 594.0, "48": 430.0, "49": 505.0, "50": 904.0, "51": 723.0, "52": 387.0, "53": 161.0, "54": 302.0, "55": 1000.0, "56": 1193.0, "57": 1071.0, "58": 1204.0, "59": 1294.0, "60": 1373.0, "61": 1345.0, "62": 1095.0, "63": 1586.0, "64": 1629.0, "65": 1685.0, "66": 1747.0, "67": 1770.0, "68": 1818.0, "69": 1092.0, "70": 1936.0, "71": 2506.0, "72": 3290.0, "73": 3695.0, "74": 3459.0, "75": 3306.0, "76": 2739.0, "77": 2041.0, "78": 1337.0, "79": 234.0, "80": 577.0, "81": 601.0, "82": 544.0, "83": 527.0, "84": 575.0, "85": 202.0, "86": NaN, "87": 973.0, "88": 1323.0, "89": 2023.0, "90": 1845.0, "91": 1408.0, "92": 917.0, "93": 913.0, "94": 1449.0, "95": 1618.0, "96": 1323.0, "97": 1173.0, "98": 1133.0, "99": 1800.0, "100": 1100.0, "101": 1100.0, "102": 1900.0, "103": NaN, "104": NaN, "105": NaN, "106": NaN, "107": NaN, "108": NaN, "109": NaN, "110": NaN, "111": NaN, "112": NaN, "113": NaN, "114": NaN, "115": NaN, "116": NaN, "117": NaN}}, "radii": {"VALUE": {"0": 0.32, "1": 0.37, "2": 1.3, "3": 0.99, "4": 0.84, "5": 0.75, "6": 0.71, "7": 0.64, "8": 0.6, "9": 0.62, "10": 1.6, "11": 1.4, "12": 1.24, "13": 1.14, "14": 1.09, "15": 1.04, "16": 1.0, "17": 1.01, "18": 2.0, "19": 1.74, "20": 1.59, "21": 1.48, "22": 1.44, "23": 1.3, "24": 1.29, "25": 1.24, "26": 1.18, "27": 1.17, "28": 1.22, "29": 1.2, "30": 1.23, "31": 1.2, "32": 1.2, "33": 1.18, "34": 1.17, "35": 1.16, "36": 2.15, "37": 1.9, "38": 1.76, "39": 1.64, "40": 1.56, "41": 1.46, "42": 1.38, "43": 1.36, "44": 1.34, "45": 1.3, "46": 1.36, "47": 1.4, "48": 1.42, "49": 1.4, "50": 1.4, "51": 1.37, "52": 1.36, "53": 1.36, "54": 2.38, "55": 2.06, "56": 1.94, "57": 1.84, "58": 1.9, "59": 1.88, "60": 1.86, "61": 1.85, "62": 1.83, "63": 1.82, "64": 1.81, "65": 1.8, "66": 1.79, "67": 1.77, "68": 1.77, "69": 1.78, "70": 1.74, "71": 1.64, "72": 1.58, "73": 1.5, "74": 1.41, "75": 1.36, "76": 1.32, "77": 1.3, "78": 1.3, "79": 1.32, "80": 1.44, "81": 1.45, "82": 1.5, "83": 1.42, "84": 1.48, "85": 1.46, "86": 2.42, "87": 2.11, "88": 2.01, "89": 1.9, "90": 1.84, "91": 1.83, "92": 1.8, "93": 1.8, "94": 1.73, "95": 1.68, "96": 1.68, "97": 1.68, "98": 1.65, "99": 1.67, "100": 1.73, "101": 1.76, "102": 1.61, "103": 1.57, "104": 1.49, "105": 1.43, "106": 1.41, "107": 1.34, "108": 1.29, "109": 1.28, "110": 1.21, "111": 1.22, "112": 1.36, "113": 1.43, "114": 1.62, "115": 1.75, "116": 1.65, "117": 1.57}, "RAD_NO": {"0": 1, "1": 2, "2": 3, "3": 4, "4": 5, "5": 6, "6": 7, "7": 8, "8": 9, "9": 10, "10": 11, "11": 12, "12": 13, "13": 14, "14": 15, "15": 16, "16": 17, "17": 18, "18": 19, "19": 20, "20": 21, "21": 22, "22": 23, "23": 24, "24": 25, "25": 26, "26": 27, "27": 28, "28": 29, "29": 30, "30": 31, "31": 32, "32": 33, "33": 34, "34": 35, "35": 36, "36": 37, "37": 38, "38": 39, "39": 40, "40": 41, "41": 42, "42": 43, "43": 44, "44": 45, "45": 46, "46": 47, "47": 48, "48": 49, "49": 50, "50": 51, "51": 52, "52": 53, "53": 54, "54": 55, "55": 56, "56": 57, "57": 58, "58": 59, "59": 60, "60": 61, "61": 62, "62": 63, "63": 64, "64": 65, "65": 66, "66": 67, "67": 68, "68": 69, "69": 70, "70": 71, "71": 72, "72": 73, "73": 74, "74": 75, "75": 76, "76": 77, "77": 78, "78": 79, "79": 80, "80": 81, "81": 82, "82": 83, "83": 84, "84": 85, "85": 86, "86": 87, "87": 88, "88": 89, "89": 90, "90": 91, "91": 92, "92": 93, "93": 94, "94": 95, "95": 96, "96": 97, "97": 98, "98": 99, "99": 100, "100": 101, "101": 102, "102": 103, "103": 104, "104": 105, "105": 106, "106": 107, "107": 108, "108": 109, "109": 110, "110": 111, "111": 112, "112": 113, "113": 114, "114": 115, "115": 116, "116": 117, "117": 118}}}
'''
    cm5_model = json.loads(mod_string)
    a0_df = pd.DataFrame.from_dict(cm5_model['A0'])
    rd_df = pd.DataFrame.from_dict(cm5_model['radii'])
    pt_df = pd.DataFrame.from_dict(cm5_model['PeriodicTable'])
    return (a0_df,rd_df,pt_df)

